#include "lexer.h"
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

bool string_slice_equal(StringSlice slice, const char* data)
{
    if (strlen(data) != slice.length) {
        return false;
    }
    return strncmp(data, slice.data, slice.length) == 0;
}

void string_construct(String* string)
{
    *string = (String) {
        .data = NULL,
        .length = 0,
        .capacity = 0,
    };
}

void string_destroy(String* string)
{
    if (string->data) {
        free(string->data);
    }
}

const size_t string_starting_alloc_size = 8;

void string_from_cstr(String* string, const char* value)
{
    string_construct(string);
    string_append_cstr(string, value);
}

void string_from_slice(String* string, StringSlice slice)
{
    string_construct(string);
    for (size_t i = 0; i < slice.length; ++i) {
        string_append_char(string, slice.data[i]);
    }
}

void string_append(String* string, const String* value)
{
    string_append_cstr(string, value->data);
}

void string_append_char(String* string, char value)
{
    if (string->length + 1 >= string->capacity) {
        if (string->capacity == 0) {
            string->capacity = string_starting_alloc_size;
        } else {
            string->capacity *= 2;
        }
        if (string->data == NULL) {
            string->data = malloc(string->capacity * sizeof(char));
            ASSERT(string->data);
        } else {
            char* new_buffer
                = realloc(string->data, string->capacity * sizeof(char));
            ASSERT(new_buffer);
            string->data = new_buffer;
        }
    }
    string->data[string->length] = value;
    string->length += 1;
    string->data[string->length] = '\0';
}

void string_append_cstr(String* string, const char* value)
{
    size_t value_length = strlen(value);

    if (string->length + value_length + 1 > string->capacity) {
        if (string->capacity == 0) {
            string->capacity = string_starting_alloc_size;
        }
        while (string->length + value_length + 1 > string->capacity) {
            string->capacity *= 2;
        }
        char* new_buffer
            = realloc(string->data, string->capacity * sizeof(char));
        ASSERT(new_buffer);
        string->data = new_buffer;
    }

    memcpy(&string->data[string->length], value, value_length);
    string->length += value_length;
    string->data[string->length] = '\0';
}

void string_append_format(String* string, const char* format, ...)
{
    va_list varargs;
    va_start(varargs, format);

    size_t format_length = (size_t)vsnprintf(NULL, 0, format, varargs);

    va_end(varargs);

    va_start(varargs, format);

    if (string->length + format_length + 1 > string->capacity) {
        if (string->capacity == 0) {
            string->capacity = string_starting_alloc_size;
        } else {
            string->capacity *= 2;
        }
        while (string->length + format_length + 1 > string->capacity) {
            string->capacity *= 2;
        }
        char* new_buffer
            = realloc(string->data, string->capacity * sizeof(char));
        ASSERT(new_buffer);
        string->data = new_buffer;
    }

    size_t written = (size_t)vsnprintf(
        &string->data[string->length],
        string->capacity - string->length,
        format,
        varargs
    );
    string->length += written;
    string->data[string->length] = '\0';

    va_end(varargs);
}

bool string_equal(const String* string, const char* value)
{
    return strncmp(value, string->data, string->length + 1);
}

void file_char_reader_construct(FileCharReader* reader, FILE* file)
{
    *reader = (FileCharReader) {
        .next = file_char_reader_next,
        .value = file_char_reader_value,
        .file = file,
        .buffer = { 0 },
    };
    string_construct(&reader->buffer);
}

void file_char_reader_destroy(FileCharReader* reader)
{
    string_destroy(&reader->buffer);
}

char file_char_reader_next(FileCharReader* reader)
{
    int read_maybe_char = fgetc(reader->file);
    if (read_maybe_char == EOF) {
        return '\0';
    }
    char read_char = (char)read_maybe_char;
    string_append_char(&reader->buffer, read_char);
    return (char)read_char;
}

StringSlice file_char_reader_value(
    const FileCharReader* reader, size_t index, size_t length
)
{
    ASSERT(index + length <= reader->buffer.length);
    return (StringSlice) {
        .data = &reader->buffer.data[index],
        .length = length,
    };
}

void error_construct(Error* error, Pos pos, String message)
{
    *error = (Error) {
        .pos = pos,
        .message = message,
    };
}

void error_destroy(Error* error) { string_destroy(&error->message); }

void error_collector_construct(ErrorCollector* collector)
{
    const size_t errors_start_capacity = 64;
    *collector = (ErrorCollector) {
        .errors = malloc(errors_start_capacity * sizeof(Error)),
        .errors_length = 0,
        .errors_capacity = errors_start_capacity,
    };
}

void error_collector_destroy(ErrorCollector* collector)
{
    for (size_t i = 0; i < collector->errors_length; ++i) {
        error_destroy(&collector->errors[i]);
    }
    free(collector->errors);
}

void error_collector_add(ErrorCollector* collector, Error error)
{
    if (collector->errors_length >= collector->errors_capacity) {
        Error* new_buffer = realloc(
            collector->errors, collector->errors_capacity * sizeof(Error)
        );
        ASSERT(new_buffer);
        collector->errors = new_buffer;
    }
    collector->errors[collector->errors_length] = error;
    collector->errors_length += 1;
}

const char* token_type_value(TokenType type)
{
    switch (type) {
        case TokenTypeEof:
            return "Eof";
        case TokenTypeError:
            return "error";
        case TokenTypeId:
            return "Id";
        case TokenTypeInt:
            return "Int";
        case TokenTypeChar:
            return "Char";
        case TokenTypeString:
            return "String";
        case TokenTypeLParen:
            return "(";
        case TokenTypeRParen:
            return ")";
        case TokenTypeLBrace:
            return "{";
        case TokenTypeRBrace:
            return "}";
        case TokenTypeLBracket:
            return "[";
        case TokenTypeRBracket:
            return "]";
    }
}

#define LEXER_ADD_ERROR(LEXER, POS, ...)                                       \
    {                                                                          \
        String error_message;                                                  \
        string_construct(&error_message);                                      \
        string_append_format(&error_message, __VA_ARGS__);                     \
        Error error;                                                           \
        error_construct(&error, (POS), error_message);                         \
        error_collector_add((LEXER)->errors, error);                           \
    }

Lexer lexer_create(CharReader* reader, ErrorCollector* errors)
{
    return (Lexer) {
        .reader = reader,
        .errors = errors,
        .current = reader->next(reader),
        .pos = { 0, 1, 1 },
    };
}

Token lexer_next(Lexer* lexer)
{
    Pos pos = lexer->pos;
    if (lexer_done(lexer)) {
        return lexer_token(lexer, TokenTypeEof, pos);
    }
    switch (lexer->current) {
        case ' ':
        case '\t':
        case '\r':
        case '\n':
            return lexer_skip_whitespace(lexer);
        case '"':
            return lexer_lex_string(lexer);
        case '(':
            return (
                lexer_step(lexer), lexer_token(lexer, TokenTypeLParen, pos)
            );
        case ')':
            return (
                lexer_step(lexer), lexer_token(lexer, TokenTypeRParen, pos)
            );
        case '{':
            return (
                lexer_step(lexer), lexer_token(lexer, TokenTypeLBrace, pos)
            );
        case '}':
            return (
                lexer_step(lexer), lexer_token(lexer, TokenTypeRBrace, pos)
            );
        case '[':
            return (
                lexer_step(lexer), lexer_token(lexer, TokenTypeLBracket, pos)
            );
        case ']':
            return (
                lexer_step(lexer), lexer_token(lexer, TokenTypeRBracket, pos)
            );
        case '/':
            return lexer_lex_slash(lexer);
        case '#':
            return lexer_lex_hashtag(lexer);
        case '0' ... '9':
            return lexer_lex_int(lexer);
        case '+':
        case '-':
        case '*':
        case '<':
        case '>':
        case '=':
        case '!':
        case 'a' ... 'z':
        case 'A' ... 'Z':
            return lexer_lex_id(lexer);
        default:
            LEXER_ADD_ERROR(
                lexer, pos, "unsupported character `%c`", lexer->current
            );
            return (lexer_step(lexer), lexer_token(lexer, TokenTypeError, pos));
    }
}

Token lexer_lex_id(Lexer* lexer)
{
    Pos pos = lexer->pos;
    lexer_step(lexer);
    while (true) {
        switch (lexer->current) {
            case '+':
            case '-':
            case '*':
            case '<':
            case '>':
            case '=':
            case '!':
            case '0' ... '9':
            case 'a' ... 'z':
            case 'A' ... 'Z':
                lexer_step(lexer);
                break;
            default:
                goto break_loop;
        }
    }
break_loop:
    return lexer_token(lexer, TokenTypeId, pos);
}

Token lexer_lex_int(Lexer* lexer)
{
    Pos pos = lexer->pos;
    lexer_step(lexer);
    while (true) {
        switch (lexer->current) {
            case '0' ... '9':
                lexer_step(lexer);
                break;
            default:
                goto break_loop;
        }
    }
break_loop:
    return lexer_token(lexer, TokenTypeInt, pos);
}

Token lexer_lex_char(Lexer* lexer)
{
    Pos pos = lexer->pos;
    lexer_step(lexer);
    if (lexer_done(lexer)) {
        LEXER_ADD_ERROR(
            lexer, pos, "malformed character literal, got unexpected Eof"
        );
        return lexer_token(lexer, TokenTypeError, pos);
    }
    if (lexer->current == '\'') {
        LEXER_ADD_ERROR(
            lexer, pos, "malformed character literal, got unexpected `'`"
        );
        return lexer_token(lexer, TokenTypeError, pos);
    }
    if (lexer->current == '\\') {
        lexer_step(lexer);
        if (lexer_done(lexer)) {
            LEXER_ADD_ERROR(
                lexer, pos, "malformed character literal, got unexpected Eof"
            );
            return lexer_token(lexer, TokenTypeError, pos);
        }
        lexer_step(lexer);
    }
    if (lexer_done(lexer)) {
        LEXER_ADD_ERROR(
            lexer, pos, "malformed character literal, expected `'`, got Eof"
        );
        return lexer_token(lexer, TokenTypeError, pos);
    }
    if (lexer->current != '\'') {
        LEXER_ADD_ERROR(
            lexer,
            pos,
            "malformed character literal, expected `'`, got `%c`",
            lexer->current
        );
        return lexer_token(lexer, TokenTypeError, pos);
    }
    lexer_step(lexer);
    return lexer_token(lexer, TokenTypeChar, pos);
}

Token lexer_lex_string(Lexer* lexer)
{
    Pos pos = lexer->pos;
    lexer_step(lexer);
    while (!lexer_done(lexer)) {
        char maybe_escape_char = lexer->current;
        lexer_step(lexer);
        if (maybe_escape_char == '\\') {
            if (lexer_done(lexer)) {
                LEXER_ADD_ERROR(
                    lexer,
                    pos,
                    "malformed string literal escape sequence, got unexpected "
                    "Eof"
                );
                return lexer_token(lexer, TokenTypeError, pos);
            }
            lexer_step(lexer);
        }
    }
    if (lexer_done(lexer)) {
        LEXER_ADD_ERROR(
            lexer, pos, "malformed string literal, expected `\"`, got Eof"
        );
        return lexer_token(lexer, TokenTypeError, pos);
    }
    if (lexer->current != '\"') {
        LEXER_ADD_ERROR(
            lexer,
            pos,
            "malformed string literal, expected `\"`, got `%c`",
            lexer->current
        );
        return lexer_token(lexer, TokenTypeError, pos);
    }
    lexer_step(lexer);
    return lexer_token(lexer, TokenTypeString, pos);
}

Token lexer_lex_slash(Lexer* lexer)
{
    Pos pos = lexer->pos;
    lexer_step(lexer);
    if (lexer->current == '*') {
        lexer_step(lexer);
        int depth = 1;
        char last = '\0';
        while (!lexer_done(lexer) && depth > 0) {
            if (last == '*' && lexer->current == '/') {
                depth -= 1;
            } else if (last == '/' && lexer->current == '*') {
                depth += 1;
            }
            last = lexer->current;
            lexer_step(lexer);
        }
        if (depth > 0) {
            String error_message;
            string_from_cstr(
                &error_message,
                "malformed multiline comment, expected `*/` before end"
            );
            Error error;
            error_construct(&error, pos, error_message);
            error_collector_add(lexer->errors, error);
            return lexer_token(lexer, TokenTypeError, pos);
        }
        return lexer_next(lexer);
    }
    if (lexer->current == '/') {
        lexer_step(lexer);
        while (!lexer_done(lexer) && lexer->current != '\n') {
            lexer_step(lexer);
        }
        return lexer_next(lexer);
    }
    return lexer_token(lexer, TokenTypeId, pos);
}

Token lexer_lex_hashtag(Lexer* lexer)
{
    lexer_step(lexer);
    while (!lexer_done(lexer) && lexer->current != '\n') {
        lexer_step(lexer);
    }
    return lexer_next(lexer);
}

Token lexer_skip_whitespace(Lexer* lexer)
{
    lexer_step(lexer);
    while (true) {
        switch (lexer->current) {
            case ' ':
            case '\t':
            case '\r':
            case '\n':
                lexer_step(lexer);
                break;
            default:
                goto break_loop;
        }
    }
break_loop:
    return lexer_next(lexer);
}

Token lexer_token(const Lexer* lexer, TokenType type, Pos pos)
{
    return (Token) {
        .type = type,
        .pos = pos,
        .length = lexer->pos.index - pos.index,
    };
}

void lexer_step(Lexer* lexer)
{
    if (lexer->current == '\n') {
        lexer->pos.line += 1;
        lexer->pos.col = 1;
    } else if (!lexer_done(lexer)) {
        lexer->pos.col += 1;
    }
    lexer->pos.index += 1;
    lexer->current = lexer->reader->next(lexer->reader);
}

bool lexer_done(const Lexer* lexer) { return lexer->current == '\0'; }
