#ifndef LEXER_H
#define LEXER_H

#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

#define PANIC(...)                                                             \
    (fprintf(stderr, "panic: "),                                               \
     fprintf(stderr, __VA_ARGS__),                                             \
     fprintf(stderr, ", at %s:%d in %s()\n", __FILE__, __LINE__, __func__),    \
     exit(1));

#define ASSERT(CONDITION)                                                      \
    {                                                                          \
        if (!(CONDITION)) {                                                    \
            (fprintf(stderr, "assertion failed: "),                            \
             fprintf(                                                          \
                 stderr,                                                       \
                 "(%s), at %s:%d in %s()\n",                                   \
                 #CONDITION,                                                   \
                 __FILE__,                                                     \
                 __LINE__,                                                     \
                 __func__                                                      \
             ),                                                                \
             exit(1));                                                         \
        }                                                                      \
    };

typedef struct {
    const char* data;
    size_t length;
} StringSlice;

bool string_slice_equal(StringSlice slice, const char* data);

typedef struct {
    char* data;
    size_t length;
    size_t capacity;
} String;

void string_construct(String* string);
void string_destroy(String* string);
void string_from_cstr(String* string, const char* value);
void string_from_slice(String* string, StringSlice slice);
void string_append(String* string, const String* value);
void string_append_char(String* string, char value);
void string_append_cstr(String* string, const char* value);
void string_append_format(String* string, const char* format, ...);
bool string_equal(const String* string, const char* value);

typedef struct CharReader CharReader;
struct CharReader {
    char (*next)(CharReader* reader);
    StringSlice (*value)(const CharReader* reader, size_t index, size_t length);
};

typedef struct FileCharReader FileCharReader;
struct FileCharReader {
    char (*next)(FileCharReader* reader);
    StringSlice (*value)(
        const FileCharReader* reader, size_t index, size_t length
    );
    FILE* file;
    String buffer;
};

void file_char_reader_construct(FileCharReader* reader, FILE* file);
void file_char_reader_destroy(FileCharReader* reader);
char file_char_reader_next(FileCharReader* reader);
StringSlice file_char_reader_value(
    const FileCharReader* reader, size_t index, size_t length
);

typedef struct {
    size_t index;
    size_t line;
    size_t col;
} Pos;

typedef struct {
    Pos pos;
    String message;
} Error;

void error_construct(Error* error, Pos pos, String message);
void error_destroy(Error* error);

typedef struct {
    Error* errors;
    size_t errors_length;
    size_t errors_capacity;
} ErrorCollector;

void error_collector_construct(ErrorCollector* collector);
void error_collector_destroy(ErrorCollector* collector);
void error_collector_add(ErrorCollector* collector, Error error);

typedef enum {
    TokenTypeEof,
    TokenTypeError,
    TokenTypeId,
    TokenTypeInt,
    TokenTypeChar,
    TokenTypeString,
    TokenTypeLParen,
    TokenTypeRParen,
    TokenTypeLBrace,
    TokenTypeRBrace,
    TokenTypeLBracket,
    TokenTypeRBracket,
} TokenType;

const char* token_type_value(TokenType type);

typedef struct {
    TokenType type;
    Pos pos;
    size_t length;
} Token;

typedef struct {
    CharReader* reader;
    ErrorCollector* errors;
    char current;
    Pos pos;
} Lexer;

Lexer lexer_create(CharReader* reader, ErrorCollector* errors);
Token lexer_next(Lexer* lexer);

Token lexer_lex_id(Lexer* lexer);
Token lexer_lex_int(Lexer* lexer);
Token lexer_lex_char(Lexer* lexer);
Token lexer_lex_string(Lexer* lexer);
Token lexer_lex_slash(Lexer* lexer);
Token lexer_lex_hashtag(Lexer* lexer);
Token lexer_skip_whitespace(Lexer* lexer);
Token lexer_token(const Lexer* lexer, TokenType type, Pos pos);
void lexer_step(Lexer* lexer);
bool lexer_done(const Lexer* lexer);

#endif
