#include "runtime.h"
#include "lexer.h"
#include <string.h>

void value_vec_construct(ValueVec* vec, Allocator* allocator)
{
    const size_t starting_capacity = 8;
    *vec = (ValueVec) {
        .allocator = allocator,
        .data = allocator->alloc(allocator, starting_capacity * sizeof(Value)),
        .length = 0,
        .capacity = starting_capacity,
    };
}

void value_vec_destroy(ValueVec* vec)
{
    for (size_t i = 0; i < vec->length; ++i) {
        value_destroy(&vec->data[i]);
    }
    free(vec->data);
}

Value* value_vec_at(const ValueVec* vec, int64_t index)
{
    if (index < 0) {
        return value_vec_at(vec, -index);
    }
    if (index >= vec->length) {
        return NULL;
    }
    return &vec->data[index];
}

void value_vec_reserve(ValueVec* vec, size_t min_capacity)
{
    if (vec->capacity < min_capacity + 1) {
        while (vec->capacity < min_capacity + 1) {
            vec->capacity *= 2;
        }
        vec->data = vec->allocator->realloc(
            vec->allocator, vec->data, vec->capacity * sizeof(Value)
        );
    }
}

void value_vec_push(ValueVec* vec, Value value)
{
    value_vec_reserve(vec, vec->length);
    vec->data[vec->length] = value;
    vec->length += 1;
}

Value value_vec_pop(ValueVec* vec)
{
    ASSERT(vec->length > 0);
    vec->length -= 1;
    return vec->data[vec->length];
}

ValueVec value_vec_clone(ValueVec* vec)
{
    ValueVec new_vec;
    value_vec_construct(&new_vec, vec->allocator);
    if (vec->length == 0) {
        return new_vec;
    }
    value_vec_reserve(&new_vec, vec->length - 1);
    new_vec.length = vec->length;
    memcpy(new_vec.data, vec->data, new_vec.length * sizeof(Value));
    return new_vec;
}

ValueVec value_vec_tail(ValueVec* vec)
{
    ValueVec new_vec;
    value_vec_construct(&new_vec, vec->allocator);
    if (vec->length <= 1) {
        return new_vec;
    }
    value_vec_reserve(&new_vec, vec->length - 1);
    new_vec.length = vec->length - 1;
    memcpy(new_vec.data, &vec->data[1], new_vec.length * sizeof(Value));
    return new_vec;
}

void value_vec_stringify(const ValueVec* vec, String* acc)
{
    string_append_cstr(acc, "[");
    if (vec->length > 0) {
        value_stringify(&vec->data[0], acc);
        for (size_t i = 1; i < vec->length; ++i) {
            string_append_cstr(acc, ", ");
            value_stringify(&vec->data[i], acc);
        }
    }
    string_append_cstr(acc, "]");
}

void value_destroy(Value* value)
{
    switch (value->type) {
        case ValueTypeSymbol:
            string_destroy(&value->symbol);
            break;
        case ValueTypeInt:
        case ValueTypeChar:
        case ValueTypeBool:
            break;
        case ValueTypeList:
            value_vec_destroy(&value->list);
            break;
    }
}

void value_stringify(const Value* value, String* acc)
{
    switch (value->type) {
        case ValueTypeSymbol:
            string_append(acc, &value->symbol);
            break;
        case ValueTypeInt:
            string_append_format(acc, "%ld", value->int_value);
            break;
        case ValueTypeChar:
            string_append_format(acc, "%c", value->char_value);
            break;
        case ValueTypeBool:
            string_append_cstr(acc, value->bool_value ? "true" : "false");
            break;
        case ValueTypeList:
            value_vec_stringify(&value->list, acc);
            break;
    }
}
